using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;

namespace TestSearch
{
    class Program
    {
        enum IndexMethod
        {
            None = 0,
            Naive,
            SubString,
            Trie,
        };
        
        struct LookupResult
        {
            public string subKey;
            public double seconds;
            public int results;
        };

        static void Main(string[] args)
        {
            if (args.Length < 4)
            {
                DisplayInstructions();
                return;
            }

            int numItems = Convert.ToInt32(args[0]);
            string filename = args[1];
            string filter = args[2];
            IndexMethod method = TextToSearchMethod(args[3]);
            if (method == IndexMethod.None)
            {
                DisplayInstructions();
            }
            
            List<string> items = ReadFileItems(filename);
            Console.WriteLine("{0:N0} total items in file", items.Count);
            //limit the number of items to how many we have
            numItems = Math.Min(numItems, items.Count);
            Console.WriteLine("\nSummary:\n\nWill index {0:N0} items in {1} and search for \"{2}\" using {3} method",
                numItems, filename, filter, method);
            
            //create the indexer and time it
            IIndexer<bool> indexer = CreateIndex(method, items, numItems);
            //do the searches and time them
            DoSearches(indexer, filter);
        }

        private static IIndexer<bool> CreateIndex(IndexMethod method, IList<string> keys, int numItems)
        {
            HighPerformanceTimer createTimer = new HighPerformanceTimer();
            long memoryUseBefore = GC.GetTotalMemory(true);
            //get the indexer we need to use
            IIndexer<bool> indexer = null;
            switch (method)
            {
                case IndexMethod.Naive:
                    indexer = new NaiveIndexer<bool>();
                    break;
                case IndexMethod.SubString:
                    indexer = new SubStringIndexer<bool>(numItems,3);
                    break;
                case IndexMethod.Trie:
                    indexer = new TrieIndexer<bool>(3);
                    break;
                default:
                    Debug.Assert(false);
                    break;
            }
            createTimer.Start();
            //add everything to the index
            for (Int32 currentItem = 0; currentItem < numItems; currentItem++)
            {
                indexer.AddItem(keys[currentItem], true, (UInt32)currentItem);
            }
            
            createTimer.Stop();
            long memoryUseAfter = GC.GetTotalMemory(true);
            long memoryDelta =  memoryUseAfter - memoryUseBefore;
            Console.WriteLine("{0} Index creation: {1:N5}s", method, createTimer.Duration);
            Console.WriteLine("Index uses {0:N0} bytes of memory", memoryDelta);
            return indexer;
        }

        private static void DoSearches(IIndexer<bool> indexer, string filter)
        {
            List<LookupResult> results = new List<LookupResult>();
            //do consecutive searches for all keys up to substring
            for (int i = 1; i <= filter.Length; i++)
            {
                string subKey = filter.Substring(0, i);
                LookupResult result = new LookupResult();
                HighPerformanceTimer searchTimer = new HighPerformanceTimer();
                searchTimer.Start();
                
                IList<bool> found = indexer.Lookup(subKey);

                searchTimer.Stop();
                //save search results
                result.seconds = searchTimer.Duration;
                result.subKey = subKey;
                result.results = found.Count;

                results.Add(result);
            }

            string template = "{0,-10}|{1,10}|{2,10}";
            Console.WriteLine("----------|----------|----------");
            Console.WriteLine(template, "Subkey", "Seconds", "# Found");
            Console.WriteLine("----------|----------|----------");
            double totalSeconds = 0.0;
            for (int i = 0; i < results.Count; i++)
            {
                Console.WriteLine(template, results[i].subKey, results[i].seconds.ToString("N5"), results[i].results.ToString("N0"));
                totalSeconds += results[i].seconds;
            }
            Console.WriteLine("----------|----------|----------");
            Console.WriteLine(template, "TOTAL", totalSeconds.ToString("N5"), "");
        }

        #region Utility functions
        private static void DisplayInstructions()
        {
            Console.WriteLine("Usage: TestSearch <#items> <inputfile> <filter> [naive|substring|trie]");
            Console.WriteLine("Example: TestSearch 10000 myfile.txt hello naive");
        }

        private static IndexMethod TextToSearchMethod(string text)
        {
            if (string.Compare(text, "naive", true) == 0)
                return IndexMethod.Naive;
            if (string.Compare(text, "substring", true) == 0)
                return IndexMethod.SubString;
            if (string.Compare(text, "trie", true) == 0)
                return IndexMethod.Trie;

            return IndexMethod.None;
        }

        private static List<string> ReadFileItems(string filename)
        {
            string[] lines = File.ReadAllLines(filename);
            return new List<string>(lines);
        }
        #endregion
    }
}
