using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

namespace TestSearch
{
    class TrieIndexer<T> : IIndexer<T>
    {
        private class IndexStruct
        {
            public UInt32 sortOrder;
            public string key;
            public T val;

            public IndexStruct(string key, T val, UInt32 sortOrder)
            {
                this.key = key;
                this.val = val;
                this.sortOrder = sortOrder;
            }
        };

        private class Trie
        {
            private char _val;
            private Dictionary<char, Trie> _next = new Dictionary<char, Trie>();
            private IList<IndexStruct> _results;

            #region Properties
            public char Value
            {
                get
                {
                    return _val;
                }
                set
                {
                    _val = value;
                }
            }

            public IList<IndexStruct> Results
            {
                get
                {
                    return _results;
                }
            }
 
            public Dictionary<char, Trie> Next
            {
                get
                {
                    return _next;
                }
            }

            #endregion

            public void AddValue(IndexStruct indexStruct)
            {
                if (_results == null)
                {
                    _results = new List<IndexStruct>();
                }
                _results.Add(indexStruct);
            }

            public void AddNextTrie(char c, Trie nextTrie)
            {
                this.Next[c] = nextTrie;
            }

            /// <summary>
            /// Initializes a new instance of the <see cref="Trie"/> class.
            /// </summary>
            /// <param name="val">The val.</param>
            public Trie(char val)
            {
                _val = val;
            }


            public IList<IndexStruct> GetCombinedResults()
            {
                Queue<IList<IndexStruct>> queue = new Queue<IList<IndexStruct>>();
                //first enqueue items in this node
                if (_results != null && _results.Count > 0)
                {
                    queue.Enqueue(_results);
                }
                
                //get items from sub-nodes
                foreach (Trie t in _next.Values)
                {
                    IList<IndexStruct> r = t.GetCombinedResults();
                    queue.Enqueue(r);
                }

                //merge all items together
                while (queue.Count > 1)
                {
                    IList<IndexStruct> a = queue.Dequeue();
                    IList<IndexStruct> b = queue.Dequeue();
                    queue.Enqueue(Merge(a, b));
                }
                return queue.Dequeue();
            }

            private IList<IndexStruct> Merge(IList<IndexStruct> a, IList<IndexStruct> b)
            {
                if (a == null || a.Count == 0)
                {
                    return b;
                }
                if (b == null || b.Count == 0)
                {
                    return a;
                }
                int iA = 0, iB = 0;
                List<IndexStruct> results = new List<IndexStruct>(a.Count + b.Count);
                while (iA < a.Count || iB < b.Count)
                {
                    if (iA < a.Count && 
                        (iB == b.Count || a[iA].sortOrder < b[iB].sortOrder))
                    {
                        results.Add(a[iA]);
                        iA++;
                    }
                    else if (iA<a.Count && iB<b.Count && 
                        a[iA].sortOrder == b[iB].sortOrder)
                    {
                        //if they're equal, make sure we skip the other one so we don't add it later
                        results.Add(a[iA]);
                        iA++;
                        iB++;
                    }
                    else
                    {
                        results.Add(b[iB]);
                        iB++;
                    }
                    
                }
                return results;
            }
        };

        private Trie _rootTrie = new Trie('\0');
        private int _maxKeyLength = 3;

        public TrieIndexer(int maxKeyLength)
        {
            _maxKeyLength = maxKeyLength;
        }

        #region IIndexer<T> Members
        /// <summary>
        /// Adds an item to the index
        /// </summary>
        /// <param name="key">The key.</param>
        /// <param name="value">The value.</param>
        /// <param name="sortOrder"></param>
        public void AddItem(string key, T value, UInt32 sortOrder)
        {
            string toAdd = key.ToLower();

            toAdd = RemoveUnneededCharacters(toAdd);
            
            IndexStruct indexStruct = new IndexStruct(toAdd, value, sortOrder);

            List<string> subStrings = GetSubStrings(toAdd);
            
            foreach (string ss in subStrings)
            {
                Trie currentTrie = _rootTrie;
                for (int i = 0; i < ss.Length; i++)
                {
                    char c = ss[i];
                    Trie nextTrie = null;
                    if (!currentTrie.Next.TryGetValue(c, out nextTrie))
                    {
                        nextTrie = new Trie(c);
                        currentTrie.AddNextTrie(c, nextTrie);
                    }
                    currentTrie = nextTrie;
                }
                currentTrie.AddValue(indexStruct);
            }
        }

        public IList<T> Lookup(string filterText)
        {
            Trie currentTrie = _rootTrie;
            
            int maxTrieLength = Math.Min(filterText.Length, _maxKeyLength);
            for (int i = 0; i < maxTrieLength; i++)
            {
                Trie nextTrie = null;
                if (currentTrie.Next.TryGetValue(filterText[i], out nextTrie))
                {
                    currentTrie = nextTrie;
                }
                else
                {
                    //no results
                    return new List<T>();
                }
            }
            Debug.Assert(currentTrie != null);
            
            return GetResults(filterText, currentTrie);
        }

        private IList<T> GetResults(string filterText, Trie trie)
        {
            List<T> results = new List<T>();

            IList<IndexStruct> preResults = trie.GetCombinedResults();
            if (preResults.Count <= 0)
            {
                return results;
            }
            results.Capacity = preResults.Count;
            uint prevSortOrder = 0;
            foreach (IndexStruct item in preResults)
            {
                Debug.Assert(item.sortOrder > prevSortOrder);
                prevSortOrder = item.sortOrder;

                if (filterText.Length <= _maxKeyLength || 
                    item.key.IndexOf(filterText, 
                        StringComparison.InvariantCultureIgnoreCase) >= 0)
                {
                    results.Add(item.val);
                }
            }
            return results;
        }

        #endregion

        /// <summary>
        /// Removes all but alphanumeric characters from the input string
        /// </summary>
        /// <param name="original">The original.</param>
        /// <returns>A string with all unneeded characters removed</returns>
        private string RemoveUnneededCharacters(string original)
        {
            char[] array = new char[original.Length];
            int destIndex = 0;
            for (int i = 0; i < original.Length; i++)
            {
                char c = original[i];
                if (char.IsLetterOrDigit(c))
                {
                    array[destIndex] = c;
                    destIndex++;
                }
            }
            return new string(array, 0, destIndex);
        }

        /// <summary>
        /// Gets a list of all indexable substrings.
        /// </summary>
        /// <param name="key">The key.</param>
        /// <returns>List of substrings</returns>
        private List<string> GetSubStrings(string key)
        {
            List<string> results = new List<string>();
            //we only need to return substrings that
            //themselves don't begin other substrings
            //easy to do--return first _maxKeyLength characters
            //or whatever's left, if shorter
            for (int start = 0; start < key.Length; start++)
            {
                int len = Math.Min(_maxKeyLength, key.Length - start);
                string sub = key.Substring(start, len);
                //remove this if() to speed up index creation at the cost of slightly longer
                //lookup time
                if (key.IndexOf(sub) == start)
                {
                    results.Add(sub);
                }
            }
            return results;
        }
    }
}
